#include <stdio.h>
#include <proto.h>

#include "gtwidget.h"
#include "gtmanager.h"
#include "gtappshell.h"
#include "gtapp.h"
#include "gtresname.h"
#include "gtcommon.h"
#include "gtconfig.h"
#include "gterror.h"

/************************************************************************\
*
* GtWidget class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define PBACK		0x0001
#define PFORE		0x0002
#define PLHT		0x0004
#define PDRK		0x0008
#define PFILL		0x0010
#define PDIS		0x0020
#define ABACK		0x0040
#define AFORE		0x0080
#define ALHT		0x0100
#define ADRK		0x0200
#define AFILL		0x0400
#define ADIS		0x0800
#define PCM			0x1000
#define PFNT		0x2000
#define PBPIX		0x4000
#define ABCOL		0x8000
#define PCURS		0x10000

#include "dpat.bm"

/************************************************************************\
*
* Create and init widget.
*
\************************************************************************/
GtWidget *GtMkWidget(char *n,GtManager *p,GtTag *tgs)
{
	GtWidget *obj;
	
	if(!(obj=new GtWidget()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtWidget object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init widget.
*
\************************************************************************/
GtWidget *GtMkWidget(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtWidget *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkWidget(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Initial setup of widget.
*
\************************************************************************/
GtWidget::GtWidget(void)
{
	app=NULL;
	dsp=NULL;
	parent=NULL;
	gfx=NULL;
	userd=NULL;
	name=NULL;
	emask=0;
	rxpos=rypos=0;
	xpos=ypos=0;
	xsize=ysize=1;
	minxs=minys=1;
	maxxs=maxys=GTS_MAX;
	flags=GTWF_UNBUILT|GTWF_ONCE;
}

/************************************************************************\
*
* Destroy widget.
*
\************************************************************************/
void GtWidget::Destroy(void)
{
	int n;
	
	dsp->RemWidget(this);
	if(parent)
		parent->RemChild(this);
	else
	{
		if(dsp->GetRoot()==this)
			dsp->SetRoot(NULL);
	}
	if(flags & GTWF_NOWIN)
		return;
	if(gfx->cmap)
	{
		for(n=0; n<GT_NR_COLS; n++)
		{
			if(gfx->pflags & (ABACK<<n))
				gfx->cmap->FreeC(gfx->cols[n]);
		}
		if(gfx->pflags & ABCOL)
			gfx->cmap->FreeC(gfx->bcol);
		if(gfx->pflags & PCM)
			GtDestroy(gfx->cmap);
	}
	if(gfx->fnt && (gfx->pflags & PFNT))
		GtDestroy(gfx->fnt);
	if(gfx->dpix)
		GtDestroy(gfx->dpix);
	for(n=0; n<GT_NR_GCS; n++)
	{
		if(gfx->gcs[n])
			GtDestroy(gfx->gcs[n]);
	}
	if(gfx->cur && (gfx->pflags & PCURS))
		GtDestroy(gfx->cur);
	if(gfx->win)
		GtDestroy(gfx->win);
	if(gfx->bpix && (gfx->pflags & PBPIX))
		GtDestroy(gfx->bpix);
	GtBase::Destroy();
}

/************************************************************************\
*
* Destroy widget if and only if it is not a child of the current widget.
*
\************************************************************************/
void GtWidget::DeInit(void)
{
	this->Map(FALSE);
	flags|=GTWF_UNBUILT;
	if(dsp->GetCurrentWidget() && (dsp->GetCurrentWidget()->IsChildOf(this) ||
		(dsp->GetCurrentWidget()==this)))
		flags|=GTWF_DESTROY;
	else
		GtBase::DeInit();
}

/************************************************************************\
*
* Setup widget.
*
\************************************************************************/
sbool GtWidget::Init(char *nm,GtManager *p,GtTag *tgs)
{
	GtTag *t,*tt;
	GtWidget *o;
	int pos=GTP_LAST,n;
	sbool map=TRUE,wless=FALSE;
	char *cstr[GT_NR_COLS],*rbcol;
	char *flist=NULL,*bbm=NULL,*bxpm=NULL;
	GtGfx g;
	XColor xc;
	
	// Name of widget
	name=nm;
	if(!name)
		name="<unnamed>";
	// Check if parent is widget or GtAppDisp object
	if(p->IsSubClassOf("GtAppDisp"))
		dsp=(GtAppDisp *)p;
	else
	{
		// Parent is widget so get GtAppDisp object from it
		if(!p->IsSubClassOf("GtManager"))
		{
			GtError(this,GTEC_NOADD,"Illegal parent!\n");
			return(FALSE);
		}
		parent=p;
		dsp=p->dsp;
	}
	if(!dsp)
	{
		GtError(this,GTEC_NODISPLAY,"No GtDisplay!\n");
		return(FALSE);
	}
	// Get GtApp object
	app=(GtApp *)dsp->GetA(GT_App,0);
	if(!app)
	{
		GtError(this,GTEC_NOAPP,"No GtApp!\n");
		return(FALSE);
	}
	// Check if parent is disabled
	if(parent && (parent->flags & GTWF_DISABLED))
		flags|=GTWF_DISABLED;
	// Initialize default color names
	cstr[GTBACK]=GTDEF_BACK;
	cstr[GTFORE]=GTDEF_FORE;
	cstr[GTLHT]=GTDEF_LIGHT;
	cstr[GTDRK]=GTDEF_DARK;
	cstr[GTFILL]=GTDEF_FILL;
	cstr[GTDIS]=GTDEF_DIS;
	rbcol=GTDEF_FORE;
	// Read resources
	app->GetRes(this,GTR_XPOS,GTRT_WORD,&xpos);
	app->GetRes(this,GTR_YPOS,GTRT_WORD,&ypos);
	app->GetRes(this,GTR_XSIZE,GTRT_WORD,&xsize);
	app->GetRes(this,GTR_YSIZE,GTRT_WORD,&ysize);
	app->GetRes(this,GTR_BACK,GTRT_STRING,&cstr[GTBACK]);
	app->GetRes(this,GTR_FORE,GTRT_STRING,&cstr[GTFORE]);
	app->GetRes(this,GTR_LHT,GTRT_STRING,&cstr[GTLHT]);
	app->GetRes(this,GTR_DRK,GTRT_STRING,&cstr[GTDRK]);
	app->GetRes(this,GTR_FILL,GTRT_STRING,&cstr[GTFILL]);
	app->GetRes(this,GTR_DIS,GTRT_STRING,&cstr[GTDIS]);
	app->GetRes(this,GTR_FLIST,GTRT_STRING,&flist);
	app->GetRes(this,GTR_BCOL,GTRT_STRING,&rbcol);
	app->GetRes(this,GTR_BXPM,GTRT_STRING,&bxpm);
	app->GetRes(this,GTR_BBM,GTRT_STRING,&bbm);
	app->GetRes(this,GTR_WLESS,GTRT_BOOL,&wless);
	memset(&g,0,sizeof(GtGfx));
	g.pflags=ABACK|AFORE|ALHT|ADRK|AFILL|ADIS;
	tt=tgs;
	while((t=GtNextTag(&tt)))
	{
		switch(t->tag)
		{
			case GT_Disabled:
				if(t->d.val)
					flags|=GTWF_DISABLED;
				else
					flags&=~GTWF_DISABLED;
				break;
			case GT_Position:
				pos=t->d.val;
				break;
			case GT_Colormap:
				g.cmap=(GtColormap *)t->d.ptr;
				break;
			case GT_XPos:
				xpos=t->d.val;
				break;
			case GT_YPos:
				ypos=t->d.val;
				break;
			case GT_XSize:
				xsize=t->d.val;
				break;
			case GT_YSize:
				ysize=t->d.val;
				break;
			case GT_Visual:
				g.vis=(Visual *)t->d.ptr;
				break;
			case GT_Depth:
				g.depth=t->d.val;
				break;
			case GT_Mapped:
				if(t->d.val)
					map=TRUE;
				else
					map=FALSE;
				break;
			case GT_Background:
				g.cols[GTBACK]=t->d.val;
				g.pflags&=~ABACK;
				break;
			case GT_Foreground:
				g.cols[GTFORE]=t->d.val;
				g.pflags&=~AFORE;
				break;
			case GT_Light:
				g.cols[GTLHT]=t->d.val;
				g.pflags&=~ALHT;
				break;
			case GT_Dark:
				g.cols[GTDRK]=t->d.val;
				g.pflags&=~ADRK;
				break;
			case GT_Fill:
				g.cols[GTFILL]=t->d.val;
				g.pflags&=~AFILL;
				break;
			case GT_Disable:
				g.cols[GTDIS]=t->d.val;
				g.pflags&=~ADIS;
				break;
			case GT_Font:
				g.fnt=(GtFont *)t->d.ptr;
				break;
			case GT_FontList:
				flist=(char *)t->d.ptr;
				break;
			case GT_BackPixmap:
				g.bpix=(GtPixmap *)t->d.ptr;
				break;
			case GT_BackXPM:
				bxpm=(char *)t->d.ptr;
				break;
			case GT_BackBitmap:
				bbm=(char *)t->d.ptr;
				break;
			case GT_BackColor:
				g.bcol=t->d.val;
				g.pflags&=~ABCOL;
				break;
			case GT_Cursor:
				g.cur=(GtCursor *)t->d.ptr;
				break;
			case GT_EnterLeave:
				if(t->d.val)
					flags|=GTWF_ENTERCB;
				else
					flags&=~GTWF_ENTERCB;
				break;
			case GT_WindowLess:
				if(t->d.val)
					wless=TRUE;
				else
					wless=FALSE;
				break;
		}
	}
	// Set event mask for widget
	this->SetEMask(0);
	// Check if this widget can be windowless
	if(wless && (((emask & ExposureMask)!=emask) || !parent))
	{
		GtError(this,GTEC_NOXRESOURCE,"Widget can't be windowless!\n");
		return(FALSE);
	}
	// Check if this widget should be windowless
	if((emask==NoEventMask) && !this->Manage())
		wless=TRUE;
	// Handle windowless widgets
	if(wless)
	{
		flags|=GTWF_NOWIN;
		gfx=parent->gfx;
		parent->SetEMask(ExposureMask);
		// Get position of this widget in it's shell
		if(parent && this->Manage())
		{
			rxpos=parent->rxpos;
			rypos=parent->rypos;
		}
	}
	else
	{
		// Get position of this widget in it's shell
		if(parent && this->Manage())
		{
			rxpos=parent->rxpos+xpos;
			rypos=parent->rypos+ypos;
		}
		// Memory for graphical data
		if(!(gfx=(GtGfx *)malloc(sizeof(GtGfx))))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for widget graphics!\n");
			return(FALSE);
		}
		*gfx=g;
		// Depth of widget window
		if(!gfx->depth)
			gfx->depth=dsp->XDefaultDepth();
		// Widget colormap
		if(!gfx->cmap)
		{
			if(parent)
				gfx->cmap=parent->gfx->cmap;
			else
			{
				gfx->cmap=GtMkColormap(dsp,
					GT_Mode,GTCM_DEFAULT,
					GT_Visual,gfx->vis,
					GT_DONE);
				gfx->pflags|=PCM;
			}
		}
		if(!gfx->cmap)
			return(FALSE);
		// Widget font
		if(!gfx->fnt)
		{
			gfx->fnt=GtMkFont(dsp,GT_FontList,flist,GT_DONE);
			gfx->pflags|=PFNT;
		}
		if(!gfx->fnt)
			return(FALSE);
		// Allocate widget colors
		for(n=0; n<GT_NR_COLS; n++)
		{
			if(gfx->pflags & (ABACK<<n))
			{
				if(!gfx->cmap->AllocNamedC(cstr[n],&xc))
				{
					GtError(this,GTEC_NOXRESOURCE,"No color: %s\n",cstr[n]);
					return(FALSE);
				}
				gfx->cols[n]=xc.pixel;
			}
		}
		// Allocate color for background pixmap
		if(gfx->pflags & ABCOL)
		{
			if(!gfx->cmap->AllocNamedC(rbcol,&xc))
			{
				GtError(this,GTEC_NOXRESOURCE,"No back bitmap color: %s\n",rbcol);
				return(FALSE);
			}
			gfx->bcol=xc.pixel;
			gfx->pflags|=ABCOL;
		}
		// Create unmapped window
		gfx->win=GtMkWindow(dsp,
			GT_Parent,(parent && !this->IsSubClassOf("GtShell") ? 
									parent->GetWin() : NULL),
			GT_Colormap,gfx->cmap,
			GT_XPos,xpos,
			GT_YPos,ypos,
			GT_XSize,xsize,
			GT_YSize,ysize,
			GT_Depth,gfx->depth,
			GT_EventMask,emask,
			GT_Cursor,gfx->cur,
			(gfx->vis ? GT_Visual : GT_IGNORE),gfx->vis,
			GT_DONE);
		if(!gfx->win)
		{
			GtError(this,GTEC_NOXRESOURCE,"No window for widget!\n");
			return(FALSE);
		}
		// Find window for shared GtGCs
		if(parent && (parent->gfx->vis==gfx->vis) &&
				(parent->gfx->depth==gfx->depth))
			gfx->shwin=parent->gfx->shwin;
		else
			gfx->shwin=gfx->win;
		// Create shared GtGCs
		for(n=0; n<GT_NR_GCS-1; n++)
		{
			gfx->gcs[n]=GtMkSharedGC(dsp,
				GT_Drawable,gfx->shwin,
				GT_Foreground,gfx->cols[n],
				GT_Background,gfx->cols[GTBACK],
				GT_Font,gfx->fnt,
				GT_GraphExpose,FALSE,
				GT_DONE);
			if(!gfx->gcs[n])
			{
				GtError(this,GTEC_NOXRESOURCE,"No GtGC for color: %d\n",n);
				return(FALSE);
			}
		}
		// Create disable pattern shared pixmap
		gfx->dpix=GtMkSharedPixmap(dsp,
			GT_Drawable,gfx->shwin,
			GT_BitmapData,dpat_bits,
			GT_XSize,dpat_width,
			GT_YSize,dpat_height,
			GT_Depth,1,
			GT_DONE);
		if(!gfx->dpix)
		{
			GtError(this,GTEC_NOXRESOURCE,"No disable pattern pixmap!\n");
			return(FALSE);
		}
		// Create shared disable GtGC
		gfx->gcs[GTDIS]=GtMkSharedGC(dsp,
			GT_Drawable,gfx->shwin,
			GT_Foreground,gfx->cols[GTDIS],
			GT_Background,gfx->cols[GTBACK],
			GT_Stipple,gfx->dpix,
			GT_FillStyle,FillStippled,
			GT_GraphExpose,FALSE,
			GT_DONE);
		if(!gfx->gcs[GTDIS])
		{
			GtError(this,GTEC_NOXRESOURCE,"No disable GtGC!\n");
			return(FALSE);
		}
		// Create background pixmap
		if(bxpm && GtXpmSupport())
		{
			gfx->bpix=GtMkSharedPixmap(dsp,
				GT_Drawable,gfx->shwin,
				GT_XpmName,bxpm,
				GT_Depth,gfx->depth,
				GT_Foreground,gfx->bcol,
				GT_Background,gfx->cols[GTBACK],
				GT_DONE);
			if(!gfx->bpix)
			{
				GtError(this,GTEC_NOXRESOURCE,"No background pixmap!\n");
				return(FALSE);
			}
			gfx->pflags|=PBPIX;
		}
		else if(bbm)
		{
			gfx->bpix=GtMkSharedPixmap(dsp,
				GT_Drawable,gfx->shwin,
				GT_BitmapName,bbm,
				GT_Depth,gfx->depth,
				GT_Foreground,gfx->bcol,
				GT_Background,gfx->cols[GTBACK],
				GT_DONE);
			if(!gfx->bpix)
			{
				GtError(this,GTEC_NOXRESOURCE,"No background pixmap!\n");
				return(FALSE);
			}
			gfx->pflags|=PBPIX;
		}
		if(gfx->bpix)
			gfx->win->SetV(GT_BackPixmap,gfx->bpix,GT_DONE);
		else
			gfx->win->SetV(GT_BackPixel,gfx->cols[GTBACK],GT_DONE);
		if(flags & GTWF_ENTERCB)
			this->SetEMask(EnterWindowMask|LeaveWindowMask);
	}
	// Set mapping flags
	this->Map(map);
	// Link widget to it's parent
	if(!parent)
	{
		if((o=dsp->GetRoot()))
		{
			if(!o->IsSubClassOf("GtManager"))
			{
				GtError(this,GTEC_NOADD,"Illegal parent: %s\n",o->GetName());
				return(FALSE);
			}
			parent=(GtManager *)o;
			if(!parent->AddChild(this,pos))
				return(FALSE);
		}
		else
			dsp->SetRoot(this);
	}
	else
	{
		if(!parent->IsSubClassOf("GtManager"))
		{
			GtError(this,GTEC_NOADD,"Illegal parent: %s\n",parent->GetName());
			return(FALSE);
		}
		if(!parent->AddChild(this,pos))
			return(FALSE);
	}
	if(parent && parent->IsManager() && this->Manage())
		parent->SetLInfo(this,tgs);
	dsp->AddWidget(this);
	return(TRUE);
}

/************************************************************************\
*
* Setup object.
*
\************************************************************************/
sbool GtWidget::InitV(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	int ps;
	sbool ret;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	ret=this->Init(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(ret);
}

/************************************************************************\
*
* Find name and parent tags and then setup normal.
*
\************************************************************************/
sbool GtWidget::Init(GtTag *tgs)
{
	char *name;
	GtManager *p;
	
	name=(char *)GtGetTagPtr(tgs,GT_Name,NULL);
	p=(GtManager *)GtGetTagPtr(tgs,GT_Parent,NULL);
	return(this->Init(name,p,tgs));
}

/************************************************************************\
*
* Set object attributes.
*
\************************************************************************/
sbool GtWidget::Set(GtTag *tgs)
{
	GtTag *t,*tt;
	GtManager *pr;
	sbool ps=FALSE,vis=FALSE,fc=FALSE;
	sbool cch[GT_NR_COLS]={FALSE};
	sbool cbpat=FALSE,cel=FALSE,ccur=FALSE,ccm=FALSE;
	int xp,yp,xs,ys,n;
	char *flist=NULL,*bbm=NULL,*bxpm=NULL;
	
	xp=xpos;
	yp=ypos;
	xs=xsize;
	ys=ysize;
	tt=tgs;
	while((t=GtNextTag(&tt)))
	{
		switch(t->tag)
		{
			case GT_Colormap:
				if(t->d.ptr && !(flags & GTWF_NOWIN))
				{
					if(gfx->cmap && (gfx->pflags & PCM))
					{
						GtDestroy(gfx->cmap);
						gfx->pflags&=~PCM;
					}
					gfx->cmap=(GtColormap *)t->d.ptr;
					ccm=TRUE;
				}
				break;
			case GT_XPos:
				xp=t->d.val;
				ps=TRUE;
				break;
			case GT_YPos:
				yp=t->d.val;
				ps=TRUE;
				break;
			case GT_XSize:
				xs=t->d.val;
				ps=TRUE;
				break;
			case GT_YSize:
				ys=t->d.val;
				ps=TRUE;
				break;
			case GT_Disabled:
				if(t->d.val)
					flags|=GTWF_DISABLED;
				else
					flags&=~GTWF_DISABLED;
				vis=TRUE;
				break;
			case GT_Background:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & ABACK)
					gfx->cmap->FreeC(gfx->cols[GTBACK]);
				gfx->cols[GTBACK]=t->d.val;
				gfx->pflags&=~ABACK;
				cch[GTBACK]=TRUE;
				vis=TRUE;
				break;
			case GT_Foreground:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & AFORE)
					gfx->cmap->FreeC(gfx->cols[GTFORE]);
				gfx->cols[GTFORE]=t->d.val;
				gfx->pflags&=~AFORE;
				cch[GTFORE]=TRUE;
				vis=TRUE;
				break;
			case GT_Light:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & ALHT)
					gfx->cmap->FreeC(gfx->cols[GTLHT]);
				gfx->cols[GTLHT]=t->d.val;
				gfx->pflags&=~ALHT;
				cch[GTLHT]=TRUE;
				vis=TRUE;
				break;
			case GT_Dark:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & ADRK)
					gfx->cmap->FreeC(gfx->cols[GTDRK]);
				gfx->cols[GTDRK]=t->d.val;
				gfx->pflags&=~ADRK;
				cch[GTDRK]=TRUE;
				vis=TRUE;
				break;
			case GT_Fill:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & AFILL)
					gfx->cmap->FreeC(gfx->cols[GTFILL]);
				gfx->cols[GTFILL]=t->d.val;
				gfx->pflags&=~AFILL;
				cch[GTFILL]=TRUE;
				vis=TRUE;
				break;
			case GT_Disable:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & ADIS)
					gfx->cmap->FreeC(gfx->cols[GTDIS]);
				gfx->cols[GTDIS]=t->d.val;
				gfx->pflags&=~ADIS;
				cch[GTDIS]=TRUE;
				vis=TRUE;
				break;
			case GT_Font:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->fnt && (gfx->pflags & PFNT))
					GtDestroy(gfx->fnt);
				gfx->pflags&=~PFNT;
				gfx->fnt=(GtFont *)t->d.ptr;
				fc=TRUE;
				break;
			case GT_FontList:
				if(flags & GTWF_NOWIN)
					break;
				flist=(char *)t->d.ptr;
				break;
			case GT_BackPixmap:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->bpix && (gfx->pflags & PBPIX))
					GtDestroy(gfx->bpix);
				gfx->pflags&=~PBPIX;
				gfx->bpix=(GtPixmap *)t->d.ptr;
				cbpat=TRUE;
				vis=TRUE;
				break;
			case GT_BackXPM:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->bpix && (gfx->pflags & PBPIX))
					GtDestroy(gfx->bpix);
				gfx->pflags&=~PBPIX;
				bxpm=(char *)t->d.ptr;
				cbpat=TRUE;
				vis=TRUE;
				break;
			case GT_BackBitmap:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->bpix && (gfx->pflags & PBPIX))
					GtDestroy(gfx->bpix);
				gfx->pflags&=~PBPIX;
				bbm=(char *)t->d.ptr;
				cbpat=TRUE;
				vis=TRUE;
				break;
			case GT_BackColor:
				if(flags & GTWF_NOWIN)
					break;
				if(gfx->pflags & ABCOL)
					gfx->cmap->FreeC(gfx->bcol);
				gfx->pflags&=~ABCOL;
				gfx->bcol=t->d.val;
				break;
			case GT_Cursor:
				if(flags & GTWF_NOWIN)
					break;
				gfx->cur=(GtCursor *)t->d.ptr;
				ccur=TRUE;
				break;
			case GT_EnterLeave:
				if(flags & GTWF_NOWIN)
					break;
				if(t->d.val)
					flags|=GTWF_ENTERCB;
				else
					flags&=~GTWF_ENTERCB;
				cel=TRUE;
				break;
		}
	}
	if(flist)
	{
		if(gfx->fnt && (gfx->pflags & PFNT))
			GtDestroy(gfx->fnt);
		if(!(gfx->fnt=GtMkFont(dsp,GT_FontList,flist,GT_DONE)))
			return(FALSE);
		gfx->pflags|=PFNT;
	}
	for(n=0; n<GT_NR_COLS; n++)
	{
		if(cch[n] || cch[GTBACK] || fc)
		{
			if(gfx->pflags & PBACK)
			{
				if(!gfx->gcs[n]->SetV(GT_Foreground,gfx->cols[n],
															GT_Background,gfx->cols[GTBACK],
															GT_Font,gfx->fnt,
															GT_DONE))
					return(FALSE);
			}
			else
			{
				if(gfx->gcs[n])
					GtDestroy(gfx->gcs[n]);
				gfx->gcs[n]=GtMkSharedGC(dsp,
					GT_Drawable,gfx->shwin,
					GT_Foreground,gfx->cols[n],
					GT_Background,gfx->cols[GTBACK],
					GT_Font,gfx->fnt,
					GT_GraphExpose,FALSE,
					GT_DONE);
				if(!gfx->gcs[n])
				{
					GtError(this,GTEC_NOXRESOURCE,"No GtGC for color: %d\n",n);
					return(FALSE);
				}			
			}
		}
	}
	if(cch[GTBACK])
	{
		if(!gfx->bpix)
		{
			if(!gfx->win->SetV(GT_BackPixel,gfx->cols[GTBACK],GT_DONE))
				return(FALSE);
		}
	}
	if(cch[GTBACK] || cch[GTDIS])
	{
		if(gfx->gcs[GTDIS])
			GtDestroy(gfx->gcs[GTDIS]);
		// Create shared disable GtGC
		gfx->gcs[GTDIS]=GtMkSharedGC(dsp,
			GT_Drawable,gfx->shwin,
			GT_Foreground,gfx->cols[GTDIS],
			GT_Background,gfx->cols[GTBACK],
			GT_Stipple,gfx->dpix,
			GT_FillStyle,FillStippled,
			GT_GraphExpose,FALSE,
			GT_DONE);
		if(!gfx->gcs[GTDIS])
		{
			GtError(this,GTEC_NOXRESOURCE,"No disable GtGC!\n");
			return(FALSE);
		}
	}
	if(cbpat)
	{
		// Create background pixmap
		if(bxpm && GtXpmSupport())
		{
			gfx->bpix=GtMkSharedPixmap(dsp,
				GT_Drawable,gfx->shwin,
				GT_XpmName,bxpm,
				GT_Depth,gfx->depth,
				GT_Foreground,gfx->bcol,
				GT_Background,gfx->cols[GTBACK],
				GT_DONE);
			if(!gfx->bpix)
			{
				GtError(this,GTEC_NOXRESOURCE,"No background pixmap!\n");
				return(FALSE);
			}
			gfx->pflags|=PBPIX;
		}
		else if(bbm)
		{
			gfx->bpix=GtMkSharedPixmap(dsp,
				GT_Drawable,gfx->shwin,
				GT_BitmapName,bbm,
				GT_Depth,gfx->depth,
				GT_Foreground,gfx->bcol,
				GT_Background,gfx->cols[GTBACK],
				GT_DONE);
			if(!gfx->bpix)
			{
				GtError(this,GTEC_NOXRESOURCE,"No background pixmap!\n");
				return(FALSE);
			}
			gfx->pflags|=PBPIX;
		}
		if(gfx->bpix)
			gfx->win->SetV(GT_BackPixmap,gfx->bpix,GT_DONE);
		if(!gfx->bpix)
			gfx->win->SetV(GT_BackPixmap,NULL,GT_BackPixel,gfx->cols[GTBACK],GT_DONE);
	}
	if(cel)
	{
		if(flags & GTWF_ENTERCB)
			this->SetEMask(EnterWindowMask|LeaveWindowMask);
		else
			this->SetEMask(0);
	}
	if(ccm)
	{
		pr=parent;
		while(pr && !pr->IsSubClassOf("GtAppShell"))
			pr=pr->GetParent();
		if(pr)
			((GtAppShell *)pr)->UpdateColormaps();
		gfx->win->SetV(GT_Colormap,gfx->cmap,GT_DONE);
	}
	if(ccur)
		gfx->win->SetV(GT_Cursor,gfx->cur,GT_DONE);
	if(ps)
		this->MoveResize(xp,yp,xs,ys);
	if(parent && parent->IsManager() && this->Manage())
	{
		if(parent->SetLInfo(this,tgs))
			parent->Relayout();
	}
	if(vis && !ps)
		this->Refresh();
	return(TRUE);
}

/************************************************************************\
*
* Get object attributes.
*
\************************************************************************/
void GtWidget::Get(GtTag *tgs)
{
	GtTag *t,*tt;
	long *ptr;
	
	tt=tgs;
	while((t=GtNextTag(&tt)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Display:
				*ptr=(long)dsp;
				break;
			case GT_Parent:
				*ptr=(long)parent;
				break;
			case GT_Name:
				*ptr=(long)name;
				break;
			case GT_Disabled:
				if(flags & GTWF_DISABLED)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_Colormap:
				*ptr=(long)gfx->cmap;
				break;
			case GT_Window:
				*ptr=(long)gfx->win;
				break;
			case GT_XPos:
				*ptr=xpos;
				break;
			case GT_YPos:
				*ptr=ypos;
				break;
			case GT_XSize:
				*ptr=xsize;
				break;
			case GT_YSize:
				*ptr=ysize;
				break;
			case GT_Mapped:
				if(flags & GTWF_MAPPED)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_Background:
				*ptr=(long)gfx->cols[GTBACK];
				break;
			case GT_Foreground:
				*ptr=(long)gfx->cols[GTFORE];
				break;
			case GT_Light:
				*ptr=(long)gfx->cols[GTLHT];
				break;
			case GT_Dark:
				*ptr=(long)gfx->cols[GTDRK];
				break;
			case GT_Fill:
				*ptr=(long)gfx->cols[GTFILL];
				break;
			case GT_Font:
				*ptr=(long)gfx->fnt;
				break;
			case GT_BackPixmap:
				*ptr=(long)gfx->bpix;
				break;
			case GT_BackColor:
				*ptr=(long)gfx->bcol;
				break;
			case GT_Cursor:
				*ptr=(long)gfx->cur;
				break;
			case GT_EnterLeave:
				if(flags & GTWF_ENTERCB)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_WindowLess:
				if(flags & GTWF_NOWIN)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
		}
	}
	if(parent && parent->IsManager() && this->Manage())
		parent->GetLInfo(this,tgs);
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtWidget::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtWidget")==0)
		return(TRUE);
	else
		return(GtBase::IsSubClassOf(cn));
}

/************************************************************************\
*
* Print error header for any widget object with class and name.
*
\************************************************************************/
void GtWidget::PrtErrHeader(void)
{
	fprintf(stderr,"Gt error in %s named '%s':\n  ",this->GetClass(),name);
}

/************************************************************************\
*
* Try to map or unmap a widget window.
*
\************************************************************************/
void GtWidget::Map(sbool map)
{
	if(parent)
	{
		if(parent->MapReq(this,map))
		{
			if(!(flags & GTWF_NOWIN) && !(flags & GTWF_UNBUILT))
				gfx->win->Map(map);
			flags=(map ? flags|GTWF_MAPPED : flags&~GTWF_MAPPED);
		}
		if(parent && parent->IsManager() && this->Manage())
			parent->Relayout();
	}
	else
	{
		if(!(flags & GTWF_NOWIN) && !(flags & GTWF_UNBUILT))
			gfx->win->Map(map);
		flags=(map ? flags|GTWF_MAPPED : flags&~GTWF_MAPPED);
	}
}

/************************************************************************\
*
* Build last parts of widgets recursive.
*
\************************************************************************/
sbool GtWidget::Build(void)
{
	if(flags & GTWF_UNBUILT)
	{
		flags&=~GTWF_UNBUILT;
		this->Map(flags & GTWF_MAPPED ? TRUE : FALSE);
	}
	return(TRUE);
}

/************************************************************************\
*
* Try to move a widget.
*
\************************************************************************/
void GtWidget::Move(word xp,word yp)
{
	if((xp!=xpos) || (yp!=ypos))
	{
		if(parent && parent->IsManager() && this->Manage())
			parent->MoveSizeReq(this,xp,yp,xsize,ysize);
		else
			this->SetPosSize(xp,yp,xsize,ysize);
	}
}

/************************************************************************\
*
* Try to resize widget.
*
\************************************************************************/
void GtWidget::Resize(word xs,word ys)
{
	if((xs!=xsize) || (ys!=ysize))
	{
		if(parent && parent->IsManager() && this->Manage())
			parent->MoveSizeReq(this,xpos,ypos,xs,ys);
		else
			this->SetPosSize(xpos,ypos,xs,ys);
	}
}

/************************************************************************\
*
* Try to move and resize widget.
*
\************************************************************************/
void GtWidget::MoveResize(word xp,word yp,word xs,word ys)
{
	if((xp!=xpos) || (yp!=ypos) || (xs!=xsize) || (ys!=ysize))
	{
		if(parent && parent->IsManager() && this->Manage())
			parent->MoveSizeReq(this,xp,yp,xs,ys);
		else
			this->SetPosSize(xp,yp,xs,ys);
	}
}

/************************************************************************\
*
* Refresh widget graphic.
*
\************************************************************************/
void GtWidget::Refresh(void)
{
	if(gfx)
	{
		if(flags & GTWF_NOWIN)
			gfx->win->Clear(xpos,ypos,xsize,ysize);
		else
			gfx->win->Clear();
		this->Draw(NULL);
		if(flags & GTWF_DISABLED)
		{
			gtdis->SetTSOrigin(rxpos%GT_DIS_SZ,rypos%GT_DIS_SZ);
			if(flags & GTWF_NOWIN)
				gtdis->DrawRectangle(gtwin,xpos,ypos,xsize,ysize,TRUE);
			else
				gtdis->DrawRectangle(gtwin,0,0,xsize,ysize,TRUE);
		}
	}
}

/************************************************************************\
*
* Handle X-events to widget.
*
\************************************************************************/
int GtWidget::Handle(XEvent *xe)
{
	GtEventCD cd;
	int ret=GTHR_UNUSED;
	
	if(!(flags & GTWF_MAPPED))
		return(GTHR_SEARCH);
	if(xe && ((xe->xany.window==gfx->win->XWin()) || (flags & GTWF_NOWIN)))
	{
		switch(xe->type)
		{
			case Expose:
			case GraphicsExpose:
			case NoExpose:
				if(this->Draw(xe))
				{
					if((flags & GTWF_DISABLED) && (xe->type==Expose) &&
							(xe->xexpose.count==0))
					{
						gtdis->SetTSOrigin(rxpos%GT_DIS_SZ,rypos%GT_DIS_SZ);
						if(flags & GTWF_NOWIN)
							gtdis->DrawRectangle(gtwin,xpos,ypos,xsize,ysize,TRUE);
						else
							gtdis->DrawRectangle(gtwin,0,0,xsize,ysize,TRUE);
					}
					ret=GTHR_USED;
				}
				break;
			case ConfigureNotify:
			case CirculateNotify:
			case CreateNotify:
			case DestroyNotify:
			case UnmapNotify:
			case MapNotify:
			case MapRequest:
			case ReparentNotify:
			case ConfigureRequest:
			case GravityNotify:
			case ResizeRequest:
			case CirculateRequest:
				if(this->Config(xe))
					ret=GTHR_USED;
				break;
			case KeyPress:
			case KeyRelease:
				if(this->Key(xe))
					ret=GTHR_USED;
				break;
			case ButtonPress:
			case ButtonRelease:
			case MotionNotify:
			case EnterNotify:
			case LeaveNotify:
				if(this->Mouse(xe))
					ret=GTHR_USED;
				if((flags & GTWF_ENTERCB) && (xe->type==EnterNotify) && 
						(xe->xcrossing.mode==NotifyNormal))
				{
					dsp->EnterLeave(this,TRUE,xe->xcrossing.state);
				}
				else if((flags & GTWF_ENTERCB) && (xe->type==LeaveNotify) && 
						(xe->xcrossing.mode==NotifyNormal))
				{
					dsp->EnterLeave(this,FALSE,xe->xcrossing.state);
				}
				break;
			default:
				if(this->Misc(xe))
					ret=GTHR_USED;
				break;
		}
		cd.used=FALSE;
		cd.xevent=xe;
		evcb.CallCB(this,&cd);
		if(cd.used)
			ret=GTHR_USED;
		return(ret);
	}
	return(GTHR_SEARCH);
}

/************************************************************************\
*
* Check if widget is child of a given widget.
*
\************************************************************************/
sbool GtWidget::IsChildOf(GtWidget *o)
{
	GtWidget *p;
	
	p=this;
	while((p=(GtWidget *)p->GetParent()))
	{
		if(p==o)
			return(TRUE);
	}
	return(FALSE);
}

/************************************************************************\
*
* Check if widget is a named widget.
*
\************************************************************************/
GtWidget *GtWidget::Find(char *n)
{
	if(strcmp(n,name)==0)
		return(this);
	else
		return(NULL);
}

/************************************************************************\
*
* Set widget event mask.
*
\************************************************************************/
void GtWidget::SetEMask(ulong em)
{
	emask=em;
	if(gfx && !(flags & GTWF_NOWIN))
		gfx->win->SetV(GT_EventMask,emask,GT_DONE);
}

/************************************************************************\
*
* Set widget minimum size. Return TRUE if any change.
*
\************************************************************************/
sbool GtWidget::SetMinMaxSize(word xs,word ys,word mxs,word mys)
{
	word oxs,oys,omxs,omys;
	
	oxs=minxs;
	oys=minys;
	omxs=maxxs;
	omys=maxys;
	minxs=xs;
	minys=ys;
	maxxs=mxs;
	maxys=mys;
	maxxs=GTMAX(minxs,maxxs);
	maxys=GTMAX(minys,maxys);
	if((oxs!=minxs) || (oys!=minys) || (omxs!=maxxs) || (omys!=maxys))
	{

		if(parent && ((oxs>minxs) || (oys>minys) || (omxs<maxxs) || (omys<maxys)))
			parent->Relayout();
		else

			this->Resize(GTMAX(minxs,GTMIN(maxxs,xsize)),
										GTMAX(minys,GTMIN(maxys,ysize)));
		return(TRUE);
	}
	return(FALSE);
}

/************************************************************************\
*
* Add function to widget callback lists.
*
\************************************************************************/
void GtWidget::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	if(tag==GT_Event)
		evcb.AddCB(cb,ud);
}

/************************************************************************\
*
* Remove function from widget callback lists.
*
\************************************************************************/
void GtWidget::RemCB(ulong tag,GtCallB cb)
{
	if(tag==GT_Event)
		evcb.RemCB(cb);
}

/************************************************************************\
*
* Remap colors and colormap recursive for widget.
*
\************************************************************************/
void GtWidget::RemapColors(uword msk,ulong bg,ulong fg,
										ulong lt,ulong dk,ulong fl,ulong ds,GtColormap *cmap)
{
	this->SetV(
		(msk & GTRCM_BACK ? GT_Background : GT_IGNORE),bg,
		(msk & GTRCM_FORE ? GT_Foreground : GT_IGNORE),fg,
		(msk & GTRCM_LIGHT ? GT_Light : GT_IGNORE),lt,
		(msk & GTRCM_DARK ? GT_Dark : GT_IGNORE),dk,
		(msk & GTRCM_FILL ? GT_Fill : GT_IGNORE),fl,
		(msk & GTRCM_DIS  ? GT_Disable : GT_IGNORE),ds,
		(msk & GTRCM_CMAP ? GT_Colormap : GT_IGNORE),cmap,
		GT_DONE);
}
													
/************************************************************************\
*
* Set position and size of widget.
*
\************************************************************************/
void GtWidget::SetPosSize(word xp,word yp,word xs,word ys)
{
	xpos=xp;
	ypos=yp;
	xsize=xs;
	ysize=ys;
	if(parent)
	{
		if(flags & GTWF_NOWIN)
		{
			rxpos=parent->rxpos;
			rypos=parent->rypos;
		}
		else
		{
			rxpos=parent->rxpos+xpos;
			rypos=parent->rypos+ypos;
		}
	}
	if(gfx && !(flags & GTWF_NOWIN))
	{
		gfx->win->MoveResize(xpos,ypos,xsize,ysize);
	}
}

/************************************************************************\
*
* Change to private GtGC for one color.
*
\************************************************************************/
sbool GtWidget::PrivateGC(ulong tag)
{
	switch(tag)
	{
		case GT_Background:
			if(gfx->pflags & PBACK)
				break;
			GtDestroy(gfx->gcs[GTBACK]);
			if(!(gfx->gcs[GTBACK]=GtMkGC(dsp,
				GT_Drawable,gfx->win,
				GT_Foreground,gfx->cols[GTBACK],
				GT_Background,gfx->cols[GTBACK],
				GT_Font,gfx->fnt,
				GT_GraphExpose,FALSE,
				GT_DONE)))
					return(FALSE);
			gfx->pflags|=PBACK;
			break;
		case GT_Foreground:
			if(gfx->pflags & PFORE)
				break;
			GtDestroy(gfx->gcs[GTFORE]);
			if(!(gfx->gcs[GTFORE]=GtMkGC(dsp,
				GT_Drawable,gfx->win,
				GT_Foreground,gfx->cols[GTFORE],
				GT_Background,gfx->cols[GTBACK],
				GT_Font,gfx->fnt,
				GT_GraphExpose,FALSE,
				GT_DONE)))
					return(FALSE);
			gfx->pflags|=PFORE;
			break;
		case GT_Light:
			if(gfx->pflags & PLHT)
				break;
			GtDestroy(gfx->gcs[GTLHT]);
			if(!(gfx->gcs[GTLHT]=GtMkGC(dsp,
				GT_Drawable,gfx->win,
				GT_Foreground,gfx->cols[GTLHT],
				GT_Background,gfx->cols[GTBACK],
				GT_Font,gfx->fnt,
				GT_GraphExpose,FALSE,
				GT_DONE)))
					return(FALSE);
			gfx->pflags|=PLHT;
			break;
		case GT_Dark:
			if(gfx->pflags & PDRK)
				break;
			GtDestroy(gfx->gcs[GTDRK]);
			if(!(gfx->gcs[GTDRK]=GtMkGC(dsp,
				GT_Drawable,gfx->win,
				GT_Foreground,gfx->cols[GTDRK],
				GT_Background,gfx->cols[GTBACK],
				GT_Font,gfx->fnt,
				GT_GraphExpose,FALSE,
				GT_DONE)))
					return(FALSE);
			gfx->pflags|=PDRK;
			break;
		case GT_Fill:
			if(gfx->pflags & PFILL)
				break;
			GtDestroy(gfx->gcs[GTFILL]);
			if(!(gfx->gcs[GTFILL]=GtMkGC(dsp,
				GT_Drawable,gfx->win,
				GT_Foreground,gfx->cols[GTFILL],
				GT_Background,gfx->cols[GTBACK],
				GT_Font,gfx->fnt,
				GT_GraphExpose,FALSE,
				GT_DONE)))
					return(FALSE);
			gfx->pflags|=PFILL;
			break;
	}
	return(TRUE);
}

